
from epyk.core.js import JsUtils


class KeyCode(object):

  def __init__(self, component, source_event=None):
    self.__events_per_source, self._component, self.source_event = {}, component, source_event or component.dom.varId

  def custom(self, rule, jsFnc, profile=False, source_event=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param rule:
    :param jsFnc:
    :param profile:
    :param source_event: String. Optional. The source component for the event
    """
    source_event = source_event or self.source_event
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    if source_event not in self.__events_per_source:
      self.__events_per_source[source_event] = {}
    self.__events_per_source[source_event].setdefault(rule, {})["content"] = jsFnc
    self.__events_per_source[source_event][rule]['profile'] = profile

  def key(self, key_code, jsFnc, profile=False, reset=False, source_event=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param key_code:
    :param jsFnc:
    :param profile:
    :param reset:
    :param source_event: String. Optional. The source component for the event
    """
    source_event = source_event or self.source_event
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    tag = "event.which == %s" % key_code
    if reset or tag not in self.__events_per_source:
      self.__events_per_source[source_event] = {}
    self.__events_per_source[source_event].setdefault(tag, {}).setdefault("content", []).extend(jsFnc)
    self.__events_per_source[source_event][tag]['profile'] = profile

  def enter(self, jsFnc, profile=False, reset=False, source_event=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFnc: Array. Tje Javascript events
    :param profile:
    :param reset: Boolean. To set if the event should be refreshed
    :param source_event: String. Optional. The source component for the event
    """
    self.key(13, jsFnc, profile, reset, source_event)

  def tab(self, jsFnc, profile=False, reset=False, source_event=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFnc: Array. Tje Javascript events
    :param profile:
    :param reset: Boolean. To set if the event should be refreshed
    :param source_event: String. Optional. The source component for the event
    """
    self.key(9, jsFnc, profile, reset, source_event)

  def backspace(self, jsFnc, profile=False, reset=False, source_event=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFnc: Array. Tje Javascript events
    :param profile:
    :param reset: Boolean. To set if the event should be refreshed
    :param source_event: String. Optional. The source component for the event
    """
    self.key(8, jsFnc, profile, reset, source_event)

  def shift_with(self, key, jsFnc, profile=False, source_event=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param key:
    :param jsFnc:
    :param profile:
    :param source_event: String. Optional. The source component for the event
    """
    self.custom("(event.shiftKey) && (event.which == %s)" % ord(key), jsFnc, profile, source_event)

  def shift(self, jsFnc, profile=False, reset=False, source_event=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFnc: Array. Tje Javascript events
    :param profile:
    :param reset: Boolean. To set if the event should be refreshed
    :param source_event: String. Optional. The source component for the event
    """
    self.key(16, jsFnc, profile, reset, source_event)

  def control(self, jsFnc, profile=False, reset=False, source_event=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFnc: Array. Tje Javascript events
    :param profile:
    :param reset: Boolean. To set if the event should be refreshed
    :param source_event: String. Optional. The source component for the event
    """
    self.key(17, jsFnc, profile, reset, source_event)

  def alt(self, jsFnc, profile=False, reset=False, source_event=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFnc: Array. Tje Javascript events
    :param profile:
    :param reset: Boolean. To set if the event should be refreshed
    :param source_event: String. Optional. The source component for the event
    """
    self.key(18, jsFnc, profile, reset, source_event)

  def space(self, jsFnc, profile=False, reset=False, source_event=None):
    """
    Description:
    -----------
    Add an event on the space key

    Attributes:
    ----------
    :param jsFnc: Array. Tje Javascript events
    :param profile:
    :param reset: Boolean. To set if the event should be refreshed
    :param source_event:
    """
    self.key(32, jsFnc, profile, reset, source_event)

  def right(self, jsFnc, profile=False, reset=False, source_event=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFnc: Array. Tje Javascript events
    :param profile:
    :param reset: Boolean. To set if the event should be refreshed
    :param source_event: String. Optional. The source component for the event
    """
    self.key(39, jsFnc, profile, reset, source_event)

  def left(self, jsFnc, profile=False, reset=False, source_event=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFnc: Array. Tje Javascript events
    :param profile:
    :param reset: Boolean. To set if the event should be refreshed
    :param source_event: String. Optional. The source component for the event
    """
    self.key(37, jsFnc, profile, reset, source_event)

  def up(self, jsFnc, profile=False, reset=False, source_event=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFnc: Array. Tje Javascript events
    :param profile:
    :param reset: Boolean. To set if the event should be refreshed
    :param source_event: String. Optional. The source component for the event
    """
    self.key(38, jsFnc, profile, reset, source_event)

  def down(self, jsFnc, profile=False, reset=False, source_event=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFnc: Array. Tje Javascript events
    :param profile:
    :param reset: Boolean. To set if the event should be refreshed
    :param source_event: String. Optional. The source component for the event
    """
    self.key(40, jsFnc, profile, reset, source_event)

  def delete(self, jsFnc, profile=False, reset=False, source_event=None):
    """
    Description:
    -----------
    Keycode 46, the sup key

    Attributes:
    ----------
    :param jsFnc: Array. Tje Javascript events
    :param profile:
    :param reset: Boolean. To set if the event should be refreshed
    :param source_event: String. Optional. The source component for the event
    """
    self.key(46, jsFnc, profile, reset, source_event)

  def escape(self, jsFnc, profile=False, reset=False, source_event=None):
    """
    Description:
    -----------
    Keycode 27, the escape key

    Attributes:
    ----------
    :param jsFnc: Array. Tje Javascript events
    :param profile:
    :param reset: Boolean. To set if the event should be refreshed
    :param source_event: String. Optional. The source component for the event
    """
    self.key(27, jsFnc, profile, reset, source_event)

  def get_event(self):
    """
    Description:
    -----------
    Return the complete definition for the key event.
    """
    event = {}
    for source, event_fncs in self.__events_per_source.items():
      event[source] = {"content": [], 'profile': False}
      for rule, jsFnc in event_fncs.items():
        event[source]["content"].append("if(%s){%s}" % (rule, JsUtils.jsConvertFncs(jsFnc['content'], toStr=True)))
    return event



"""

"""

from epyk.core.css import Defaults_css
from epyk.core.css.styles.attributes import Attrs


class AttrIcon(Attrs):
  def __init__(self, rptObj):
    super(AttrIcon, self).__init__(rptObj)
    self.font_size = Defaults_css.font()
    self.vertical_align = "middle"
    self.display = "inline-block"
    self.margin = "auto 0"
    self.padding = "auto 0"




from epyk.core.html import geo


class ChartJs(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "ChartJs"

  def choropleths(self):
    return Choropleth(self.parent)


class Choropleth(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "ChartJs"

  def usa(self, record=None, y_columns=None, x_axis=None, title=None, profile=None, options=None, width=(100, "%"),
           height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoChartJs.Choropleth`

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    geo_chart = geo.GeoChartJs.Choropleth(self.parent.context.rptObj, width, height, htmlCode, options or {}, profile)
    return geo_chart

  def world_map(self, record=None, y_columns=None, x_axis=None, title=None, profile=None, options=None, width=(100, "%"),
           height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoChartJs.Choropleth`

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param title:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    geo_chart = geo.GeoChartJs.Choropleth(self.parent.context.rptObj, width, height, htmlCode, options or {}, profile)
    return geo_chart




from epyk.core.js.objects import JsNodeDomRect

from epyk.core.js.primitives import JsObject
from epyk.core.js.primitives import JsString
from epyk.core.js.primitives import JsNumber
from epyk.core.js.primitives import JsBoolean
from epyk.core.js.primitives import JsArray

from epyk.core.js.fncs import JsFncs

from epyk.core.js import JsUtils


class IntersectionObserverEntry(object):

  @property
  def boundingClientRect(self):
    """
    Description:
    ------------
    The IntersectionObserverEntry interface's read-only boundingClientRect property returns a DOMRectReadOnly which in essence describes a rectangle describing the smallest rectangle that contains the entire target element.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/IntersectionObserverEntry/boundingClientRect
    """

  @property
  def intersectionRatio(self):
    """
    Description:
    ------------
    The IntersectionObserverEntry interface's read-only intersectionRatio property tells you how much of the target element is currently visible within the root's intersection ratio, as a value between 0.0 and 1.0.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/IntersectionObserverEntry/intersectionRatio
    """

  @property
  def intersectionRect(self):
    """
    Description:
    ------------
    The IntersectionObserverEntry interface's read-only intersectionRect property is a DOMRectReadOnly object which describes the smallest rectangle that contains the entire portion of the target element which is currently visible within the intersection root.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/IntersectionObserverEntry/intersectionRect
    """

  @property
  def isIntersecting(self):
    """
    Description:
    ------------
    The IntersectionObserverEntry interface's read-only isIntersecting property is a Boolean value which is true if the target element intersects with the intersection observer's root.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/IntersectionObserverEntry/isIntersecting
    """

  @property
  def rootBounds(self):
    """
    Description:
    ------------
    The IntersectionObserverEntry interface's read-only rootBounds property is a DOMRectReadOnly corresponding to the target's root intersection rectangle, offset by the IntersectionObserver.rootMargin if one is specified.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/IntersectionObserverEntry/rootBounds
    """

  @property
  def target(self):
    """
    Description:
    ------------
    The IntersectionObserverEntry interface's read-only target property indicates which targeted Element has changed its amount of intersection with the intersection root.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/IntersectionObserverEntry/target
    """

  @property
  def time(self):
    """
    Description:
    ------------
    The IntersectionObserverEntry interface's read-only time property is a DOMHighResTimeStamp that indicates the time at which the intersection change occurred relative to the time at which the document was created.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/IntersectionObserverEntry/time
    """
    return JsNumber.JsNumber("%s.time")


class IntersectionObserver(object):

  def __init__(self):
    pass

  @property
  def root(self):
    """
    Description:
    ------------
    The IntersectionObserver interface's read-only root property identifies the Element whose bounds are treated as the bounding box of the viewport for the element which is the observer's target.
    If the root is null, then the bounds of the actual document viewport are used.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/IntersectionObserver/root
    """
    return JsObject.JsObject("%s.root", isPyData=False)

  @property
  def rootMargin(self):
    """
    Description:
    ------------
    The IntersectionObserver interface's read-only rootMargin property is a string with syntax similar to that of the CSS margin property.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/IntersectionObserver/rootMargin
    """
    return JsString.JsString("%s.rootMargin", isPyData=False)



  @property
  def thresholds(self):
    """
    Description:
    ------------
    The IntersectionObserver interface's read-only thresholds property returns the list of intersection thresholds that was specified when the observer was instantiated with IntersectionObserver().
    If only one threshold ratio was provided when instanitating the object, this will be an array containing that single value.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/IntersectionObserver/thresholds
    """
    return JsArray.JsArray("%s.thresholds", isPyData=False)

  def observe(self, targetElement):
    """
    Description:
    ------------
    The IntersectionObserver method observe() adds an element to the set of target elements being watched by the IntersectionObserver.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/IntersectionObserver/observe

    Attributes:
    ----------
    :param targetElement: An element whose visibility within the root is to be monitored.
                          This element must be a descendant of the root element (or contained wtihin the current document, if the root is the document's viewport).
    :return:
    """
    return

  def disconnect(self):
    """
    Description:
    ------------
    The IntersectionObserver method disconnect() stops watching all of its target elements for visibility changes.

    https://developer.mozilla.org/en-US/docs/Web/API/IntersectionObserver/disconnect
    """
    return JsFncs.JsFunction("%s.disconnect()")

  def takeRecords(self):
    """
    Description:
    ------------
    The IntersectionObserver method takeRecords() returns an array of IntersectionObserverEntry objects, one for each targeted element which has experienced an intersection change since the last time the intersections were checked, either explicitly through a call to this method or implicitly by an automatic call to the observer's callback.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/IntersectionObserver/takeRecords

    :return: An array of IntersectionObserverEntry objects, one for each target element whose intersection with the root has changed since the last time the intersections were checked.
    """
    return JsFncs.JsFunction("%s.takeRecords()")

  def unobserve(self, targetElement):
    """
    Description:
    ------------
    The IntersectionObserver method unobserve() instructs the IntersectionObserver to stop observing the specified target element.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/IntersectionObserver/unobserve

    Attributes:
    ----------
    :param targetElement: The Element to cease observing. If the specified element isn't being observed, this method does nothing and no exception is thrown.
    """
    return JsFncs.JsFunction("%s.unobserve()")




from epyk.core.html import graph


class Billboard(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "C3"

  def line(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a line chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.LineChart

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.bb.y(record, y_columns, x_axis)
    line_chart = graph.GraphBillboard.ChartLine(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def line_range(self, record, y_columns=None, x_axis=None, range=5, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a line range chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.LineChart

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.bb.y(record, y_columns, x_axis)
    line_chart = graph.GraphBillboard.ChartLine(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart._type = "area-line-range"
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def bubble(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a bubble chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.BubbleChart

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.bb.y(record, y_columns, x_axis)
    bubble_chart = graph.GraphBillboard.ChartBubble(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    bubble_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      bubble_chart.add_dataset(data['series'][i], d)
      bubble_chart.data.labels = True
    return bubble_chart

  def radar(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a radar chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.RadarChart

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.bb.y(record, y_columns, x_axis)
    radar_chart = graph.GraphBillboard.ChartRadar(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    radar_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      radar_chart.add_dataset(data['series'][i], d)
      radar_chart.data.labels = True
    return radar_chart

  def spline(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a spline chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.SplineChart

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.bb.y(record, y_columns, x_axis)
    line_chart = graph.GraphBillboard.ChartSpline(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def step(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a step chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.StepChart

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.bb.y(record, y_columns, x_axis)
    line_chart = graph.GraphBillboard.ChartSpline(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart._type = 'step'
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def area(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a area chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.AreaChart

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.bb.y(record, y_columns, x_axis)
    line_chart = graph.GraphBillboard.ChartArea(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def area_step(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a area step chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.AreaChart

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.bb.y(record, y_columns, x_axis)
    line_chart = graph.GraphBillboard.ChartArea(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart._type = "area-step"
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def timeseries(self, record, y_columns=None, x_axis=None, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------
    Display a timeseries chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.LineChart

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    line = self.line(record, y_columns, x_axis, profile, width, height, options, htmlCode)
    line.axis.x.type = "timeseries"
    line.axis.x.tick.format = "%Y-%m-%d"
    return line

  def bar(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a bar chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.BarChart

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.bb.y(record, y_columns, x_axis)
    line_chart = graph.GraphBillboard.ChartBar(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def stacked(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a stacked bar chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.StackedBarChart

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.bb.y(record, y_columns, x_axis)
    line_chart = graph.GraphBillboard.ChartBar(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(data['labels'])
    line_chart.data.groups = [data['series']]
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def hbar(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a horizontal bar chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.BarChart

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    h_bar = self.bar(record, y_columns, x_axis, profile, width, height, options, htmlCode)
    h_bar.axis.rotated = True
    return h_bar

  def scatter(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a scatter chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.ScatterPlot

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.bb.y(record, y_columns, x_axis)
    line_chart = graph.GraphBillboard.ChartScatter(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      line_chart.add_dataset(data['series'][i], d)
    return line_chart

  def pie(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a pie chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.PieChart

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.bb.y(record, y_columns, x_axis)
    pie_chart = graph.GraphBillboard.ChartPie(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    pie_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      pie_chart.add_dataset(data['series'][i], d)
    return pie_chart

  def donut(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    Description:
    ------------
    Display a donut chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.DonutChart

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis})
    data = self.parent.context.rptObj.data.bb.y(record, y_columns, x_axis)
    pie_chart = graph.GraphBillboard.ChartDonut(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    pie_chart.labels(data['labels'])
    for i, d in enumerate(data['datasets']):
      pie_chart.add_dataset(data['series'][i], d)
    return pie_chart

  def gauge(self, value, text="", profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------
    Display a gauge chart from Billboard

    Related Pages:

      https://naver.github.io/billboard.js/demo/#Chart.GaugeChart

    Attributes:
    ----------
    :param value: List of dict. The Python recordset
    :param text: List. The columns corresponding to keys in the dictionaries in the record
    :param profile:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param options:
    :param htmlCode:
    """
    g_chart = graph.GraphBillboard.ChartGauge(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    g_chart.add_dataset(text, value)
    return g_chart




from epyk.core.css.catalogs import Catalog

from epyk.core.css.styles.classes import CssStylesPivot, CssStylesTableExcel, CssStylesTable, CssStylesTabulator, CssStylesDiv, CssStylesTableAgGrid


class CatalogTable(Catalog.CatalogGroup):

  def table(self):
    """  """
    return self._set_class(CssStylesTable.CssTableBasic)

  def new_row(self):
    """  """
    return self._set_class(CssStylesTable.CssTableNewRow)

  def selected(self):
    """  """
    return self._set_class(CssStylesTable.CssTableSelected)

  def cell_comment(self):
    """  """
    return self._set_class(CssStylesTable.CssCellComment)

  def cell_save(self):
    """  """
    return self._set_class(CssStylesTable.CssCellSave)

  def editable(self):
    """  """
    return self._set_class(CssStylesTable.CssTdEditor)

  def row_details(self):
    """  """
    return self._set_class(CssStylesTable.CssTdDetails)

  def row_details_shown(self):
    """  """
    return self._set_class(CssStylesTable.CssTdDetailsShown)

  def grid_headers(self):
    """  """
    return self._set_class(CssStylesTable.CssTdGridHeaderCols)

  def grid_no_header(self):
    """  """
    return self._set_class(CssStylesTable.CssTdGridNoHeaderCols)

  def grid_row_header(self):
    """  """
    return self._set_class(CssStylesTable.CssTdGridHeaderRows)

  def grid_vals(self):
    """  """
    return self._set_class(CssStylesTable.CssTdGridVals)

  def pivot_head(self):
    """  """
    return self._set_class(CssStylesPivot.CssPivotHead)

  def datatable(self):
    """  """
    return self._set_class(CssStylesTable.CssDataTable)

  def datatable_header(self):
    """  """
    return self._set_class(CssStylesTable.CssDataTableHeader)

  def datatable_footer(self):
    """  """
    return self._set_class(CssStylesTable.CssDataTableFooter)

  def datatable_even(self):
    """  """
    return self._set_class(CssStylesTable.CssDataTableEven)

  def datatable_odd(self):
    """  """
    return self._set_class(CssStylesTable.CssDataTableOdd)

  def pivot_cell(self):
    """  """
    return self._set_class(CssStylesPivot.CssPivotCells)

  def pivot_axis(self):
    """  """
    return self._set_class(CssStylesPivot.CssPivotAxis)

  def pivot_filter_box(self):
    """  """
    return self._set_class(CssStylesPivot.CssPivotFilterBox)

  def pivot_filter_val(self):
    """  """
    return self._set_class(CssStylesPivot.CssPivotFilterVals)

  def pivot_filter_label(self):
    """  """
    return self._set_class(CssStylesPivot.CssPivotLabel)

  def pivot_filter_popup(self):
    """  """
    return self._set_class(CssStylesPivot.CssPivotFilterBoxPopUp)

  def excel(self):
    """  """
    return self._set_class(CssStylesTableExcel.CssTableExcel)

  def excel_header_cell(self):
    """  """
    return self._set_class(CssStylesTableExcel.CssTableExcelHeaderCell)

  def excel_row(self):
    """  """
    return self._set_class(CssStylesTableExcel.CssTableExcelTd)

  def excel_cell(self):
    """  """
    return self._set_class(CssStylesTableExcel.CssTableExcelCell)

  def excel_title(self):
    """  """
    return self._set_class(CssStylesTableExcel.CssTableExcelTitle)

  def excel_selected_row(self):
    """  """
    return self._set_class(CssStylesTableExcel.CssTableExcelSelectedRow)

  def excel_selected_cell(self):
    """  """
    return self._set_class(CssStylesTableExcel.CssTableExcelSelected)

  def excel_red_cell(self):
    """  """
    return self._set_class(CssStylesTableExcel.CssTableRedCells)

  def excel_red_cell_background(self):
    """  """
    return self._set_class(CssStylesTableExcel.CssTableBackGroundRedCells)

  def tabulator(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulator)

  def tabulator_footer(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulatorFooter)

  def tabulator_footer_pagination(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulatorFooterPagination)

  def tabulator_header(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulatorHeader)

  def tabulator_headers(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulatorHeaders)

  def tabulator_selected(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulatorSelected)

  def tabulator_col_content(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulatorColContent)

  def tabulator_col(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulatorCol)

  def tabulator_groups(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulatorGroups)

  def tabulator_even_rows(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulatorEvenRow)

  def tabulator_odd_rows(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulatorOddRow)

  def tabulator_rows(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulatorRow)

  def tabulator_cell(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulatorCell)

  def tabulator_tree_control(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulatorTreeControl)

  def tabulator_tree_control_expand(self):
    """  """
    return self._set_class(CssStylesTabulator.CssTabulatorTreeControlExpand)

  def table_content(self):
    """ """
    return self._set_class(CssStylesDiv.CssDivTableContent)

  def ag_head(self):
    """  """
    return self._set_class(CssStylesTableAgGrid.CssAgHead)

  def ag_row_odd(self):
    """  """
    return self._set_class(CssStylesTableAgGrid.CssAgOddRow)

  def ag_row_even(self):
    """  """
    return self._set_class(CssStylesTableAgGrid.CssAgEvenRow)

  def ag_cell_focus(self):
    """  """
    return self._set_class(CssStylesTableAgGrid.CssAgCellFocus)

  def ag_cell(self):
    """  """
    return self._set_class(CssStylesTableAgGrid.CssAgCell)





from epyk.core.html.options import Options
from epyk.core.js.packages import packageImport


class OptionsNews(Options):

  @property
  def dated(self):
    """
    Description:
    ------------
    Check default value for radio and check lists
    """
    return self._config_get(True)

  @dated.setter
  def dated(self, bool):
    if bool:
      self._report.jsImports.add('moment')
    self._config(bool)

  @property
  def classes(self):
    """
    Description:
    ------------
    Check default value for radio and check lists
    """
    return self._config_get([])

  @classes.setter
  def classes(self, class_names):
    self._config(class_names)

  @property
  def markdown(self):
    """
    Description:
    ------------
    Showdown is a Javascript Markdown to HTML converter, based on the original works by John Gruber.
    Showdown can be used client side (in the browser) or server side (with NodeJs).

    Related Pages:

      https://github.com/showdownjs/showdown
    """
    return self._config_get(False, 'showdown')

  @markdown.setter
  @packageImport("showdown")
  def markdown(self, values):
    values = {} if values is True else values
    self._config(values, 'showdown')


class OptionsAlert(Options):

  @property
  def time(self):
    """
    Description:
    ------------
    """
    return self._config_get(1000)

  @time.setter
  def time(self, attrs):
    self._config(attrs)

  @property
  def delay(self):
    """
    Description:
    ------------
    """
    return self._config_get(1000)

  @delay.setter
  def delay(self, attrs):
    self._config(attrs)

  @property
  def close(self):
    """
    Description:
    ------------
    """
    return self._config_get(True)

  @close.setter
  def close(self, bool):
    self._config(bool)

  @property
  def type(self):
    """
    Description:
    ------------
    """
    return self._config_get(None)

  @type.setter
  def type(self, value):
    self._config(value)

  @property
  def markdown(self):
    """
    Description:
    ------------
    Showdown is a Javascript Markdown to HTML converter, based on the original works by John Gruber.
    Showdown can be used client side (in the browser) or server side (with NodeJs).

    Related Pages:

      https://github.com/showdownjs/showdown
    """
    return self._config_get(False, 'showdown')

  @markdown.setter
  @packageImport("showdown")
  def markdown(self, values):
    values = {} if values is True else values
    self._config(values, 'showdown')


class OptionsChat(Options):

  @property
  def dated(self):
    """
    Description:
    ------------
    Check default value for radio and check lists
    """
    return self._config_get(True)

  @dated.setter
  def dated(self, bool):
    if bool:
      self._report.jsImports.add('moment')
    self._config(bool)

  @property
  def readonly(self):
    """
    Description:
    ------------
    Check default value for radio and check lists
    """
    return self._config_get(True)

  @readonly.setter
  def readonly(self, bool):
    self._config(bool)

  @property
  def markdown(self):
    """
    Description:
    ------------
    Showdown is a Javascript Markdown to HTML converter, based on the original works by John Gruber.
    Showdown can be used client side (in the browser) or server side (with NodeJs).

    Related Pages:

      https://github.com/showdownjs/showdown
    """
    return self._config_get(False, 'showdown')

  @markdown.setter
  @packageImport("showdown")
  def markdown(self, values):
    values = {} if values is True else values
    self._config(values, 'showdown')


class OptionFiles(Options):

  @property
  def extensions(self):
    """
    Description:
    ------------
    List all the allowed file extensions
    """
    return self._config_get(None)

  @extensions.setter
  def extensions(self, values):
    self._config(values)




import json

from epyk.core.js import JsUtils
from epyk.core.html import Defaults
from epyk.core.js.statements import JsIf
from epyk.core.js.fncs import JsFncs

from epyk.core.js.objects import JsNodeDom
from epyk.core.js.primitives import JsBoolean
from epyk.core.js.primitives import JsObjects
from epyk.core.js.packages import JsQuery
from epyk.core.js.packages import JsD3
from epyk.core.js.packages import JsQueryUi
from epyk.core.js.packages import JsCrossFilter
from epyk.core.js.packages import packageImport


class FmtNumber(object):

  def __init__(self, report, selector, value):
    self._report, self._val = report, value
    self.selector = selector

  def toFixed(self, value=None):
    """
    The toFixed() method converts a number into a string, keeping a specified number of decimals.

    https://www.w3schools.com/jsref/jsref_tofixed.asp

    :param value: Integer.
    """
    if value is None:
      return JsObjects.JsObjects.get("%s = %s.toFixed()" % (self.selector, self._val))

    return JsObjects.JsObjects.get("%s = %s.toFixed(%s)" % (self.selector, self._val, value))

  def toPrecision(self, value=None):
    """
    The toPrecision() method formats a number to a specified length.

    https://www.w3schools.com/jsref/jsref_toprecision.asp

    :param value: Integer.
    """
    if value is None:
      return JsObjects.JsObjects.get("%s = %s.toPrecision()" % (self.selector, self._val))

    return JsObjects.JsObjects.get("%s = %s.toPrecision(%s)" % (self.selector, self._val, value))

  def toExponential(self):
    """
    The toExponential() method converts a number into an exponential notation.

    https://www.w3schools.com/jsref/jsref_toexponential.asp
    """
    return JsObjects.JsObjects.get("%s = %s.toExponential()" % (self.selector, self._val))


class Formatters(object):

  def __init__(self, report, selector):
    self._report = report
    self.selector = selector

  @property
  def number(self):
    """

    https://www.w3schools.com/jsref/jsref_obj_number.asp

    """
    return FmtNumber(self._report, self.selector, "parseFloat(%s)" % self.selector)

  @packageImport("accounting")
  def toNumber(self, digit=0, thousand_sep="."):
    """

    https://openexchangerates.github.io/accounting.js/

    :param digit:
    :param thousand_sep:
    """
    thousand_sep = JsUtils.jsConvertData(thousand_sep, None)
    return JsObjects.JsObjects.get("%s = accounting.formatNumber(%s, %s, %s)" % (self.selector, self.selector, digit, thousand_sep))

  @packageImport("accounting")
  def toMoney(self, symbol="", digit=0, thousand_sep=".", decimal_sep=","):
    """

    https://openexchangerates.github.io/accounting.js/

    :param symbol:
    :param digit:
    :param thousand_sep:
    :param decimal_sep:
    """
    symbol = JsUtils.jsConvertData(symbol, None)
    thousand_sep = JsUtils.jsConvertData(thousand_sep, None)
    decimal_sep = JsUtils.jsConvertData(decimal_sep, None)
    return JsObjects.JsObjects.get("%s = accounting.formatMoney(%s, %s, %s, %s, %s)" % (self.selector, self.selector, symbol, digit, thousand_sep, decimal_sep))


class ContentFormatters(object):

  def __init__(self, report, selector):
    self._report = report
    self.selector = selector

  def __ge__(self, obj):
    return "%s >= %s" % (self.selector, JsUtils.jsConvertData(obj, None))

  def __gt__(self, obj):
    return "%s > %s" % (self.selector, JsUtils.jsConvertData(obj, None))

  def __le__(self, obj):
    return "%s <= %s" % (self.selector, JsUtils.jsConvertData(obj, None))

  def __lt__(self, obj):
    return "%s < %s" % (self.selector, JsUtils.jsConvertData(obj, None))

  def __eq__(self, obj):
    return "%s == %s" % (self.selector, JsUtils.jsConvertData(obj, None))

  def __ne__(self, obj):
    return "%s <> %s" % (self.selector, JsUtils.jsConvertData(obj, None))

  @packageImport("showdown")
  def fromMarkdown(self, options=None):
    """
    Description:
    ------------
    Convert markwdown to HTML string

    Usage::

      t.dom.content.fromMarkdown()

    Related Pages:

			https://github.com/showdownjs/showdown

    Attributes:
    ----------
    :param options: Dictionary. Options allowed in the showdown module
    """
    options = JsUtils.jsConvertData(options or {}, None)
    return JsObjects.JsObjects.get(
      "%s = (function(){ var conv = new showdown.Converter(%s); return conv.makeHtml(%s)})()" % (
        self.selector, options, self.selector))

  @packageImport("accounting")
  def toNumber(self, digit=0, thousand_sep="."):
    """
    Description:
    ------------

    Usage::

      Related Pages:

			https://openexchangerates.github.io/accounting.js/

    Attributes:
    ----------
    :param digit:
    :param thousand_sep:
    """
    thousand_sep = JsUtils.jsConvertData(thousand_sep, None)
    return JsObjects.JsObjects.get("accounting.formatNumber(%s, %s, %s)" % (self.selector, digit, thousand_sep))

  @packageImport("accounting")
  def toMoney(self, symbol="", digit=0, thousand_sep=".", decimal_sep=","):
    """
    Description:
    ------------

    Related Pages:

			https://openexchangerates.github.io/accounting.js/

    Attributes:
    ----------
    :param symbol:
    :param digit:
    :param thousand_sep:
    :param decimal_sep:
    """
    symbol = JsUtils.jsConvertData(symbol, None)
    thousand_sep = JsUtils.jsConvertData(thousand_sep, None)
    decimal_sep = JsUtils.jsConvertData(decimal_sep, None)
    return JsObjects.JsObjects.get("accounting.formatMoney(%s,%s, %s, %s, %s)" % (self.selector, symbol, digit, thousand_sep, decimal_sep))

  @property
  def number(self):
    """
    Description:
    ------------

    """
    return JsObjects.JsNumber.JsNumber("parseFloat(%s)" % self.selector)

  @property
  def string(self):
    """
    Description:
    ------------

    """
    return JsObjects.JsString.JsString("String(%s)" % self.selector)

  @property
  def date(self):
    """
    Description:
    ------------

    """
    return JsObjects.JsDate.JsDate("new Date(%s)" % self.selector)

  def toStr(self):
    return self.selector


class JsHtml(JsNodeDom.JsDoms):
  display_value = "inline-block"

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    self.htmlCode = varName if varName is not None else htmlObj.htmlCode
    self.varName, self.varData, self.__var_def = "document.getElementById('%s')" % self.htmlCode, "", None
    self._src, self._report = htmlObj, report
    self._js = []
    self._jquery, self._jquery_ui, self._d3 = None, None, None

  @property
  def val(self):
    """
    Description:
    -----------

    :return:
    """
    return JsObjects.JsObjects.get("{%s: {value: %s, timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (self.htmlCode, self.content.toStr()))

  @property
  def by_name(self):
    """
    Description:
    -----------

    :rtype: JsNodeDom.JsDomsList
    """
    if self._src.attr.get('name') is not None:
      return JsNodeDom.JsDomsList(None, "document.getElementsByName('%s')" % self._src.attr.get('name'), report=self._report)

    return self

  @property
  def inViewPort(self):
    """
    Description:
    -----------

    """
    bool = JsBoolean.JsBoolean("!(rect.bottom < 0 || rect.top - viewHeight >= 0)", varName="visibleFlag", setVar=True, isPyData=False)
    bool._js.insert(0, self._report.js.viewHeight.setVar('viewHeight'))
    bool._js.insert(0, self.getBoundingClientRect().setVar("rect"))
    return JsFncs.JsAnonymous(bool.r).return_("visibleFlag").call()

  @property
  def content(self):
    """
    Description:
    -----------

    :return:
    """
    return ContentFormatters(self._report, "%s.value" % self.varName)

  def empty(self): return '%s.value = ""' % self.varName

  @property
  def events(self):
    """
    Description:
    -----------

    :rtype: JsNodeDom.JsDomEvents
    """
    return JsNodeDom.JsDomEvents(self._src)

  @property
  def jquery(self):
    """
    Description:
    -----------

    :rtype: JsQuery.JQuery
    """
    if self._jquery is None:
      self._jquery = JsQuery.JQuery(src=self._src, selector=JsQuery.decorate_var("#%s" % self._src.htmlCode), setVar=False)
    return self._jquery

  @property
  def d3(self):
    """
    Description:
    -----------

    :rtype: JsD3.D3Select
    """
    if self._d3 is None:
      self._d3 = JsD3.D3Select(src=self._src, selector="d3.select('#%s')" % self._src.htmlCode)
    return self._d3

  @property
  def jquery_ui(self):
    """
    Description:
    -----------

    :rtype: JsQuery.JQuery
    """
    if self._jquery_ui is None:
      self._jquery_ui = JsQueryUi.JQueryUI(self._src, selector=JsQuery.decorate_var("#%s" % self._src.htmlCode), setVar=False)
    return self._jquery_ui

  @property
  def objects(self):
    """
    Description:
    -----------
    Interface to the main Javascript Classes and Primitives

    Related Pages:
:return:
    """
    return JsObjects.JsObjects(self)

  @property
  def crossfilter(self):
    """
    Description:
    -----------
    Interface to CrossFilter package

    Related Pages:

			https://github.com/square/crossfilter/wiki/API-Reference#group_all

    :return:
    """
    return JsCrossFilter.CrossFilter

  @property
  def format(self):
    """
    Specific formatters for the HTML components
    """
    return Formatters(self._report, self.content.toStr())

  def style(self, attrs):
    """
    Description:
    -----------
    Style property to change from the javascript the CSS attributes of an HTML object.

    Usage::

      button.js.style({"backgroundColor": 'red'})
    button.js.style({"backgroundColor": None})

    Related Pages:

			https://www.w3.org/TR/DOM-Level-2-Style/css.html#CSS-CSSStyleRule-style

    Attributes:
    ----------
    :param attrs:
    """
    styles = []
    for k, v in attrs.items():
      styles.append("this.style.%s = %s" % (k, json.dumps(v)))
    return ";".join(styles)

  def registerFunction(self, fncName, jsFncs, pmts=None):
    """
    Description:
    -----------
    Javascript Framework extension

    Register a predefined Javascript function
    This is only dedicated to specific Javascript transformation functions

    Usage::

      Attributes:
    ----------
    :param fncName: The function name
    :param jsFncs: The Javascript function definition

    :return: The JsObject
    """
    jsData = JsUtils.jsConvertFncs(jsFncs)
    self._src._props.setdefault('js', {}).setdefault('functions', {})[fncName] = {'content': ";".join(jsData), 'pmt': pmts}
    return self

  def hide(self):
    """
    Description:
    -----------

    Usage::

      input.js.hide()

    Related Pages:

			https://gomakethings.com/how-to-show-and-hide-elements-with-vanilla-javascript/
    """
    return self.css("display", "none")

  def show(self, inline=True, duration=None):
    """
    Description:
    -----------

    Usage::

      input.js.show()

    Related Pages:

			https://gomakethings.com/how-to-show-and-hide-elements-with-vanilla-javascript/

    Attributes:
    ----------
    :param inline: String
    :param duration: Integer. A time in second for the component display
    """
    if duration is not None:
      return super(JsHtml, self).show('inline-block' if inline else self.display_value, duration)

    return JsUtils.jsConvertData(self.css("display", 'inline-block' if inline else self.display_value), None)

  def select(self):
    """
    Description:
    -----------
    Select the content of the HTMl component
    """
    return JsObjects.JsObjects.get("%s.select()" % self.varName)

  def toggle(self, attr="display", jsVal1=None, jsVal2="none"):
    """
    Description:
    ------------

    Usage::

      input.js.toggle()
    input.js.toggle("background", "red", "blue")

    Related Pages:

			https://gomakethings.com/how-to-show-and-hide-elements-with-vanilla-javascript/

    Attributes:
    ----------
    :param attr:
    :param jsVal1:
    :param jsVal2:

    :return: A Javascript if statement
    """
    if attr == 'display' and jsVal1 is None:
      jsVal1 = self.display_value
    return JsIf.JsIf(self.css(attr) == jsVal2, [self.css(attr, jsVal1)]).else_([self.css(attr, jsVal2)])

  def highlight(self, css_attrs=None, time_event=1000):
    """
    Description:
    ------------

    Usage::

      s.dom.highlight()
    s.dom.highlight(css_attrs={"background": "red"}),

    Attributes:
    ----------
    :param css_attrs: A dictionary with the CSS attributes
    :param time_event: Integer. The time of the event
    """
    if css_attrs is None:
      css_attrs, css_attrs_origin = {}, {}
      for k, v in Defaults.HTML_HIGHLIGHT.items():
        if isinstance(v, dict):
          dyn_attrs, dyn_attrs_orign = {}, {}
          if 'color' in v:
            dyn_attrs['color'] = getattr(self._report.theme, *v['color'])
            dyn_attrs_orign['color'] = self._report.theme.greys[0]
          css_attrs[k] = v['attr'] % dyn_attrs
          css_attrs_origin[k] = self._src.attr[k] if k in self._src.attr else v['attr'] % dyn_attrs_orign
        else:
          css_attrs[k] = v
          css_attrs_origin[k] = self._src.attr[k] if k in self._src.attr else "none"
    else:
      css_attrs_origin = {}
      for k in css_attrs.keys():
        if k in self._src.attr:
          css_attrs_origin[k] = self._src.attr[k]
        else:
          css_attrs_origin[k] = "none"
    return '''%s; setTimeout(function(){%s}, %s)
      ''' % (self.css(css_attrs).r, self.css(css_attrs_origin).r, time_event)

  def loadHtml(self, htmlObjs, append=False):
    """
    Description:
    ------------
    Load during a Javascript event a component within the one using this method.
    This cannot be tested during the Python execution and should be tested in the browser.

    Tip: It is possible to add a break point to debug in the browser by adding

    Usage::

      d = rptObj.ui.div().css({"border": "1px solid black"})
    b = rptObj.ui.button("test")
    b.click([
      rptObj.js.console.debugger,
      d.dom.loadHtml(rptObj.ui.texts.label("test label").css({"color": 'blue', 'float': 'none'}))
    ])

    Attributes:
    ----------
    :param htmlObjs: List. The different HTML objects to be added to the component
    :param append: Boolean. Mention if the component should replace or append the data

    :return: The Javascript string to be added to the page
    """
    if not isinstance(htmlObjs, list):
      htmlObjs = [htmlObjs]

    jsFncs = []
    for i, h in enumerate(htmlObjs):
      h.options.managed = False
      jsFncs.append(self._report.js.objects.new(str(h), isPyData=True, varName="obj_%s" % i))
      jsFncs.append(self.innerHTML(self._report.js.objects.get("obj_%s" % i), append=append).r)
    return JsUtils.jsConvertFncs(jsFncs, toStr=True)

  def options(self, options=None):
    """
    Description:
    ------------
    Return the builder options used to generate the object on the Javascript side.
    This is not necessarily the same object than the component options as some can be only used on the Python side.

    This will not change the original option object used during the first obhect creation.
    
    Attributes:
    ----------
    :param options: Dictionary. Optional. The value to be changed
    """
    opt = dict(self._src._jsStyles)
    if options is not None:
      opt.update(options)
    return opt


class JsHtmlRich(JsHtml):

  @property
  def val(self):
    """
    Description:
    -----------

    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s, timestamp: Date.now(), offset: new Date().getTimezoneOffset()} }" % (self.htmlCode, self.content.toStr()))

  @property
  def content(self):
    return ContentFormatters(self._report, "%s.innerHTML" % self.varName)

  def select(self):
    """
    Description:
    -----------
    Select the content of the HTMl component
    """
    return JsObjects.JsObjects.get('''
     (function(node){
      if (document.body.createTextRange) {const range = document.body.createTextRange(); range.moveToElementText(node); range.select();
      } else if (window.getSelection) {
        const selection = window.getSelection(); const range = document.createRange(); range.selectNodeContents(node);
        selection.removeAllRanges(); selection.addRange(range); }
     }(%s))''' % self.varName)

  def append(self, value, new_line=True, options=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param value:
    :param options:
    """
    value = JsUtils.jsConvertData(value, None)
    if options is not None and options.get('showdown') is not None:
      self._report.jsImports.add("showdown")
      value = '''(function(d){ var conv = new showdown.Converter(%s); 
                    let frag = document.createRange().createContextualFragment(conv.makeHtml(d)); 
                    frag.firstChild.style.display = 'inline-block';frag.firstChild.style.margin = 0;  
                    return frag.firstChild.outerHTML})(%s)''' % (json.dumps({}), value)
    if new_line:
      return JsObjects.JsObjects.get("%s.innerHTML += (%s+'<br />')" % (self.htmlCode, value))
      #return JsObjects.JsObjects.get("%s.innerHTML += (%s+'\\r\\n')" % (self.htmlCode, value))

    return JsObjects.JsObjects.get("%s.innerHTML += %s)" % (self.htmlCode, value))

  def empty(self):
    """
    Description:
    ------------

    """
    return '%s.innerHTML = ""' % self.varName


class JsHtmlButton(JsHtml):

  @property
  def val(self):
    """
    Description:
    -----------

    :return:
    """
    return JsObjects.JsObjects.get('''{%s: {value: %s.innerHTML, timestamp: Date.now(), 
      offset: new Date().getTimezoneOffset(), locked: %s === 'true', name: %s}}
      ''' % (self.htmlCode, self.varName, self.getAttribute('data-locked'), self.getAttribute('name')))

  @property
  def content(self):
    return ContentFormatters(self._report, "%s.innerHTML" % self.varName)

  def loading(self, flag, multiple=False):
    """
    Description:
    -----------
    Add a loading icon to the button

    Example
    b = rptObj.ui.button("test")
    b.click([
      b.dom.loading(True),
      rptObj.js.window.setTimeout([
        b.dom.loading(False)
      ], 5000),
    ])

    :param flag:
    :param multiple:
    :return:
    """
    i_loading = '<i class="fas fa-spinner fa-spin" style="margin-right:5px"></i>'
    fnc = self.disable(False) if multiple else self.disable(flag)
    if flag:
      fnc.append("%s.innerHTML = '%s' + %s.innerHTML" % (self.varName, i_loading, self.varName))
    else:
      fnc.append("%s.innerHTML = %s.innerHTML.replace('%s', '')" % (self.varName, self.varName, i_loading))
    return fnc

  def error(self, time, color="red"):
    """
    Description:
    -----------

    :param time:
    :param color:
    :return:
    """
    return JsFncs.JsFunction("var bgColor = %s.style.borderColor; %s.style.borderColor = '%s'; setTimeout(function() {%s.style.borderColor = bgColor}, %s)" % (self.varName, self.varName, color, self.varName, time))

  def disable(self, bool=True):
    bool = JsUtils.jsConvertData(bool, None)
    return JsFncs.JsFunctions("%s.disabled = %s" % (self.varName, bool))

  def release(self, by_name=False):
    """
    Description:
    -----------

    :param by_name:
    :return:
    """
    if by_name:
      fncs = JsFncs.JsFunctions(self.by_name.css("color", ''))
      fncs.append(self.by_name.css("background-color", ''))
      fncs.append(self.by_name.css("cursor", "pointer"))
      fncs.append(self.by_name.attr('data-locked', False))
    else:
      fncs = JsFncs.JsFunctions(self.css("color", ''))
      fncs.append(self.css("background-color", ''))
      fncs.append(self.css("cursor", "pointer"))
      fncs.append(self.attr('data-locked', False))
    return fncs

  def lock(self, not_allowed=True):
    """
    Description:
    -----------

    :param lock:
    """
    fncs = JsFncs.JsFunctions(self.css("color", self.getComputedStyle('color')))
    fncs.append(self.css("background-color", self.getComputedStyle('background-color')))
    if not_allowed:
      fncs.append(self.css("cursor", "not-allowed"))
      fncs.append(self.attr('data-locked', True))
    else:
      fncs.append(self.css("cursor", "default"))
      fncs.append(self.attr('data-locked', True))
    return fncs

  def empty(self): return '%s.innerHTML = ""' % self.varName


class JsHtmlButtonMenu(JsHtmlButton):

  @property
  def val(self):
    """
    Description:
    -----------

    :return:
    """
    return JsObjects.JsObjects.get('''
        {%s: {value: %s, timestamp: Date.now(), offset: new Date().getTimezoneOffset(), label: %s.innerHTML, name: %s}}
        ''' % (self.htmlCode, self.content.toStr(), self._src.label.dom.varName, self.getAttribute('name')))

  @property
  def content(self):
    return ContentFormatters(self._report, "%s.querySelector('i').classList.contains('fa-check')" % self.varName)


class JsHtmlIcon(JsHtml):

  @property
  def val(self):
    """

    :return:
    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s, timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (self.htmlCode, self._src.dom.getAttribute("class")))

  @property
  def content(self):
    return self._src.dom.getAttribute("class")


class JsHtmlList(JsHtml):

  @property
  def val(self):
    """

    :return:
    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s.querySelector('[data-select=true]').innerHTML, timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (self.htmlCode, self.varName))

  @property
  def content(self):
    return self._src.dom.getAttribute("class")


class JsHtmlBackground(JsHtml):

  @property
  def val(self):
    """

    :return:
    """
    return JsObjects.JsObjects.get("{%s: {value: %s, timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (
      self.htmlCode, self.content.toStr()))

  @property
  def content(self):
    return ContentFormatters(self._report, self._src.dom.querySelector("div").css("backgroundColor").toStr())



"""

https://github.com/nicolaskruchten/pivottable/wiki/Parameters
"""

from epyk.core.html.options import Options


class OptionsPivot(Options):
  component_properies = ('aggregator', 'aggregatorName', 'showUI', 'rowOrder', 'colOrder', 'derivedAttributes',
                         'dataClass', 'filter', 'sorters', 'rendererOptions')

  @property
  def aggregator(self):
    """
    Description:
    ------------
    constructor for an object which will aggregate results per cell (see documentation)

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get("$.pivotUtilities.aggregators['Count']()")

  @aggregator.setter
  def aggregator(self, attrs):
    self.js_type['aggregator'] = True
    self._config(attrs)

  @property
  def aggregatorName(self):
    """
    Description:
    ------------
    Name of the aggregator, used for display purposes in some renderers

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get("Count")

  @aggregatorName.setter
  def aggregatorName(self, attrs):
    self._config(attrs)

  @property
  def showUI(self):
    """
    Description:
    ------------
    Show the drag and drop UI panel

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get(True)

  @showUI.setter
  def showUI(self, attrs):
    self._config(attrs)

  @property
  def rowOrder(self):
    """
    Description:
    ------------
    The order in which row data is provided to the renderer, must be one of "key_a_to_z", "value_a_to_z", "value_z_to_a", ordering by value orders by row total

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get("")

  @rowOrder.setter
  def rowOrder(self, attrs):
    self._config(attrs)

  @property
  def colOrder(self):
    """
    Description:
    ------------
    the order in which column data is provided to the renderer, must be one of "key_a_to_z", "value_a_to_z", "value_z_to_a", ordering by value orders by column total

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get("")

  @colOrder.setter
  def colOrder(self, attrs):
    self._config(attrs)

  @property
  def derivedAttributes(self):
    """
    Description:
    ------------
    object to define derived attributes

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get({})

  @derivedAttributes.setter
  def derivedAttributes(self, attrs):
    self._config(attrs)

  @property
  def dataClass(self):
    """
    Description:
    ------------
    Constructor for the data class to be built and passed to the Renderer (should be a subclass of the default)

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get("$.pivotUtilities.PivotData")

  @dataClass.setter
  def dataClass(self, attrs):
    self.js_type['dataClass'] = True
    self._config(attrs)

  @property
  def filter(self):
    """
    Description:
    ------------
    Called on each record, returns false if the record is to be excluded from the input before rendering or true otherwise

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get()

  @filter.setter
  def filter(self, attrs):
    self.js_type['filter'] = True
    self._config(attrs)

  @property
  def sorters(self):
    """
    Description:
    ------------
    accessed or called with an attribute name and can return a function which can be used as an argument to array.sort for output purposes.
    If no function is returned, the default sorting mechanism is a built-in "natural sort" implementation.
    Useful for sorting attributes like month names, see example 1 and example 2.

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get()

  @sorters.setter
  def sorters(self, attrs):
    self._config(attrs)

  @property
  def rendererOptions(self):
    """
    Description:
    ------------
    Object passed through to renderer as options. See Renderers and Optional Extra Renderers for details.

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get()

  @rendererOptions.setter
  def rendererOptions(self, attrs):
    self._config(attrs)


class OptionsPivotUI(OptionsPivot):
  component_properies = ('aggregator', 'aggregatorName', 'showUI', 'rowOrder', 'colOrder', 'derivedAttributes',
    'dataClass', 'filter', 'sorters', 'rendererOptions',
    'inclusions', 'exclusions', 'hiddenAttributes', 'hiddenFromAggregators', 'hiddenFromDragDrop',
    'onRefresh', 'menuLimit', 'autoSortUnusedAttrs', 'unusedAttrsVertical', 'rendererOptions')

  @property
  def inclusions(self):
    """
    Description:
    ------------
    object whose keys are attribute names and values are arrays of attribute values which denote records to include in rendering; used to prepopulate the filter menus that appear on double-click (overrides exclusions below)

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get({})

  @inclusions.setter
  def inclusions(self, attrs):
    self._config(attrs)

  @property
  def exclusions(self):
    """
    Description:
    ------------
    object whose keys are attribute names and values are arrays of attribute values which denote records to exclude from rendering; used to prepopulate the filter menus that appear on double-click

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get({})

  @exclusions.setter
  def exclusions(self, attrs):
    self._config(attrs)

  @property
  def hiddenAttributes(self):
    """
    Description:
    ------------
    object whose keys are attribute names and values are arrays of attribute values which denote records to exclude from rendering; used to prepopulate the filter menus that appear on double-click

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get({})

  @hiddenAttributes.setter
  def hiddenAttributes(self, attrs):
    self._config(attrs)

  @property
  def hiddenFromAggregators(self):
    """
    Description:
    ------------
    contains attribute names to omit from the aggregator arguments dropdowns

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get({})

  @hiddenFromAggregators.setter
  def hiddenFromAggregators(self, attrs):
    self._config(attrs)

  @property
  def hiddenFromDragDrop(self):
    """
    Description:
    ------------
    contains attribute names to omit from the drag'n'drop portion of the UI

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get({})

  @hiddenFromDragDrop.setter
  def hiddenFromDragDrop(self, attrs):
    self._config(attrs)

  @property
  def onRefresh(self):
    """
    Description:
    ------------
    called upon renderer refresh with an object representing the current UI settings (see example)

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get({})

  @onRefresh.setter
  def onRefresh(self, attrs):
    self.js_type['onRefresh'] = True
    self._config(attrs)

  @property
  def menuLimit(self):
    """
    Description:
    ------------
    maximum number of values to list in the double-click menu

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get({})

  @menuLimit.setter
  def menuLimit(self, attrs):
    self._config(attrs)

  @property
  def autoSortUnusedAttrs(self):
    """
    Description:
    ------------
    controls whether or not unused attributes are kept sorted in the UI

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get(False)

  @autoSortUnusedAttrs.setter
  def autoSortUnusedAttrs(self, attrs):
    self._config(attrs)

  @property
  def unusedAttrsVertical(self):
    """
    Description:
    ------------
    Controls whether or not unused attributes are shown vertically instead of the default which is horizontally. true means always vertical, false means always horizontal. If set to a number (as is the default) then if the attributes' names' combined length in characters exceeds the number then the attributes will be shown vertically

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get(85)

  @unusedAttrsVertical.setter
  def unusedAttrsVertical(self, bool):
    self._config(bool)

  @property
  def rendererOptions(self):
    """
    Description:
    ------------
    passed through to renderer as options. See Renderers and Optional Extra Renderers for details.

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get({})

  @rendererOptions.setter
  def rendererOptions(self, bool):
    self._config(bool)

  @property
  def renderers(self):
    """
    Description:
    ------------

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get("$.pivotUtilities.renderers")

  @renderers.setter
  def renderers(self, value):
    self.js_type['renderers'] = True
    self._config(value)

  @property
  def rendererName(self):
    """
    Description:
    ------------

    Related Pages:

      https://github.com/nicolaskruchten/pivottable/wiki/Parameters
    """
    return self._config_get()

  @rendererName.setter
  def rendererName(self, value):
    self._config(value)




from epyk.core.css.styles import GrpCls
from epyk.core.css.styles.attributes import AttrClsImage


class ClassIcon(GrpCls.ClassHtml):
  @property
  def css(self):
    """
    Description:
    ------------
    Property to the underlying CSS definition to be added to the style HTML tag of a component

    :rtype: AttrClsImage.AttrIcon
    """
    if self._css_struct is None:
      self._css_struct = AttrClsImage.AttrIcon(self.htmlObj)
    return self._css_struct





from epyk.core.js import JsUtils
from epyk.core.js.packages import JsPackage
from epyk.core.js.primitives import JsObjects


class CrossFilter(JsPackage):
  lib_alias = {'js': "crossfilter"}

  def __init__(self, src, varName, data, setVar=True):
    super(CrossFilter, self).__init__(src=src, varName=varName, selector="crossfilter(%s)" % data, setVar=setVar)

  @staticmethod
  def permute(array, index):
    """
    Description:
    -----------
    Returns a permutation of the specified array using the specified index.
    The returned array contains the corresponding element in array for each index in index, in order.
    For example, permute(["a", "b", "c"], [1, 2, 0]) returns ["b", "c", "a"].
    It is acceptable for the array and index to be different lengths, and for indexes to be duplicated or omitted

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param array:
    :param index:
    """
    return JsObjects.JsArray.JsArray("permute(%s, %s)" % (JsUtils.jsConvertData(array, None), index))

  def add(self, records):
    """
    Description:
    -----------
    Adds the specified records to this crossfilter.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param records:
    """
    return self.fnc_closure("add(%s)" % JsUtils.jsConvertData(records, None))

  def removeByValue(self, column, value):
    """
    Description:
    -----------
    Removes all records that match the current filters from this crossfilter.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param column: String. The column name in the underlying data
    :param value: Object. The value
    """
    return self.fnc_closure("remove(function (d,i) { return d[%s] === %s;})" % (JsUtils.jsConvertData(column, None), JsUtils.jsConvertData(value, None)))

  def size(self):
    """
    Description:
    -----------
    Returns the number of records in the crossfilter, independent of any filters.
    For example, if you only added a single batch of records to the Crossfilter, this method would return records.length.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference
    """
    return JsObjects.JsNumber.JsNumber("%s.size()" % self.varId)

  def groupAll(self, varName):
    """
    Description:
    -----------
    A convenience function for grouping all records and reducing to a single value. See groupAll for details.
    Note: unlike a dimension's groupAll, this grouping observes all current filters.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param varName: String. The Javascript variable name
    """
    groupObj = GroupAll(selector="%s.groupAll()" % self.varId, varName=varName, setVar=True)
    return groupObj

  def dimension(self, columns, varName=None):
    """
    Description:
    -----------
    Constructs a new dimension using the specified value accessor function

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param columns: String. The column name on which the dimension will be defined
    :param varName: String. The Javascript variable name
    """
    ools = {}
    if varName is None:
      return Dimension(selector=self.toStr(), setVar=False)

    if not isinstance(columns, list):
      columns = [(columns, int)]

    if len(columns) == 1:
      js_columns = "d[%s]" % JsUtils.jsConvertData(columns[0][0], None)
      ools[columns[0][0]] = 0
    else:
      js_frg = []
      for i, col_def in enumerate(columns):
        ools[col_def[0]] = i
        if col_def[1] == str:
          js_frg.append("d[%s]" % JsUtils.jsConvertData(col_def[0], None))
        else:
          js_frg.append("+d[%s]" % JsUtils.jsConvertData(col_def[0], None))
      js_columns = "[%s]" % ", ".join(js_frg)
    dim = Dimension(varName=varName, selector="%s.dimension(function(d) { return %s })" % (self.varId, js_columns), setVar=True)
    dim.cols = ools
    return dim


class Bissect(JsPackage):

  def by(self, value):
    """
    Description:
    -----------
    Constructs a new bisector using the specified value accessor function, which must return a naturally-ordered value.

    Attributes:
    ----------
    :param value:
    """
    raise Exception("Not implemented yet !")

  @property
  def right(self):
    """
    Description:
    -----------
    Similar to bisect.left, but returns an insertion point which comes after (to the right of) any existing entries of value in array.

    """
    raise Exception("Not implemented yet !")

  @property
  def left(self):
    """

    """
    raise Exception("Not implemented yet !")


class Heap(JsPackage):

  def byColumn(self, name):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param value:
    """
    return self.fnc("function(d) { return d['%s']; }" % name)


class Heapselect(JsPackage):

  def byColumn(self, name):
    """
    Description:
    -----------
    Constructs a new heapselect function using the specified value accessor function, which must return a naturally-ordered value.
    For example, to create a heapselect function for objects based on their property

    Attributes:
    ----------
    :param name:
    """
    return self.fnc("function(d) { return d['%s']; }" % name)


class Insertionsort(JsPackage):

  def byColumn(self, name):
    """
    Description:
    -----------
    Constructs a new insertionsort function using the specified value accessor function, which must return a naturally-ordered value.
    For example, to create a insertionsort function for objects based on their property

    Attributes:
    ----------
    :param name:
    """
    return self.fnc("function(d) { return d['%s']; }" % name)


class Quicksort(JsPackage):

  def byColumn(self, name):
    """
    Description:
    -----------
    Constructs a new quicksort function using the specified value accessor function, which must return a naturally-ordered value.
    For example, to create a quicksort function for objects based on their property

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param name: String. The column name
    """
    return self.fnc("function(d) { return d['%s']; }" % name)


class Dimension(JsPackage):

  def filter(self, jsData):
    """
    Description:
    -----------
    Filters records such that this dimension's value matches value, and returns this dimension.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference
    """
    self._js.append("filter(%s)" % JsUtils.jsConvertData(jsData, None))
    return self

  def filterExact(value):
    """
    Description:
    -----------
    Filters records such that this dimension's value equals value, and returns this dimension. For example:

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param value:
    """

  def filterRange(self, min, max):
    """
    Description:
    -----------
    Filters records such that this dimension's value is greater than or equal to range[0], and less than range[1], returning this dimension.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param min:
    :param max:
    """
    self._js.append("filterRange([%s, %s])" % (min, max))
    return self

  def filterOnColumn(self, value, column):
    """
    Description:
    -----------

    :param column_index:
    :param value:
    """
    if column is None:
      return JsObjects.JsObject.JsObject("%(id)s.filter(function(d) { return d === %(value)s} )" % {'id': self.varId, 'value': JsUtils.jsConvertData(value, None)})

    return JsObjects.JsObject.JsObject("%(id)s.filter(function(d) { return d[%(column)s] === %(value)s} )" % {'id': self.varId, 'column': self.cols[column], 'value': JsUtils.jsConvertData(value, None)})

  def filterFunction(function):
    """
    Description:
    -----------
    ilters records such that the specified function returns truthy when called with this dimension's value, and returns this dimension.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param function:
    """

  def filterAll(self):
    """
    Description:
    -----------
    Clears any filters on this dimension, selecting all records and returning this dimension. For example:

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference
    """
    return self.fnc("filterAll()")

  def id(self):
    """
    Description:
    -----------
    Returns the numeric id of the dimension. For use with crossfilter.isElementFiltered.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference#dimension_group
    """
    return JsObjects.JsNumber.JsNumber("%s.id()" % self.varId)

  def top(self, k=None):
    """
    Description:
    -----------
    Returns a new array containing the top k records, according to the natural order of this dimension.
    The returned array is sorted by descending natural order.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param k: The number of entries to keep

    :return: An array with the data
    """
    if k is None:
      return JsObjects.JsArray.JsArray("%s.top(Infinity)" % self.varId)

    return JsObjects.JsArray.JsArray("%s.top(%s)" % (self.varId, k))

  def bottom(self, k):
    """
    Description:
    -----------
    Returns a new array containing the bottom k records, according to the natural order of this dimension.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param k: The number of entries to keep

    :return: An array with the data
    """
    return JsObjects.JsArray.JsArray("%s.bottom(%s)" % (self.varId, k))

  def dispose(self):
    """
    Description:
    -----------
    Removes this dimension (and its groups) from its crossfilter.
    This frees up space for other dimensions to be added to this crossfilter.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference
    """
    return self.fnc_closure("dispose()")

  def group(self, varName):
    """
    Description:
    -----------
    Constructs a new grouping for the given dimension, according to the specified groupValue function, which takes a dimension value as input and returns the corresponding rounded value.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference
    """
    groupObj = Group(selector="%s.group()" % self.varId, varName=varName, setVar=True)
    return groupObj

  def GroupAll(self, varName):
    """
    Description:
    -----------
    Constructs a new grouping for the given dimension, according to the specified groupValue function, which takes a dimension value as input and returns the corresponding rounded value.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference
    """
    groupObj = GroupAll(selector="%s.groupAll()" % self.varId, varName=varName, setVar=True)
    return groupObj

  def groupFunction(self, varName, Fnc):
    groupObj = Group(selector="%s.group(%s)" % (self.varId, Fnc), varName=varName, setVar=True)
    return groupObj

    # groupObj = Group("%s.group(%s)" % (self.toStr(), jsFnc))
    # groupObj._set_var = self._set_var
    # return groupObj

  def hasCurrentFilter(self):
    """
    Description:
    -----------
    Returns truthy if a filter has been set, or falsy if a filter has not been set on this dimension.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference
    """
    return JsObjects.JsBoolean.JsBoolean.get("%s.hasCurrentFilter()" % self.varId)

  def quicksort(self, varName):
    """
    Description:
    -----------
    Sorts the specified subset of the array in-place, returning the array; the lower bound lo is an inclusive index, and the upper bound hi is an exclusive index.
    To sort the entire array, specify a lo of 0 and a hi of array.length.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference#dimension_group

    Attributes:
    ----------
    :param varName:
    """
    return Quicksort(selector="%s.quicksort" % self.varId, varName=varName, setVar=True)

  def insertionsort(self, varName):
    """
    Description:
    -----------
    Sorts the specified subset of the array in-place, returning the array; the lower bound lo is an inclusive index, and the upper bound hi is an exclusive index.
    To sort the entire array, specify a lo of 0 and a hi of array.length.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference#dimension_group

    Attributes:
    ----------
    :param varName:
    """
    return Insertionsort(selector="%s.insertionsort" % self.varId, varName=varName, setVar=True)

  def heapselect(self, varName):
    """
    Description:
    -----------
    The identity heapselect function; suitable for numbers, dates, strings, and other naturally-comparable objects.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference#dimension_group

    Attributes:
    ----------
    :param varName:
    """
    return Heapselect(selector="%s.heapselect" % self.varId, varName=varName, setVar=True)

  def heap(self, varName):
    """
    Description:
    -----------
    The identity heapselect function; suitable for numbers, dates, strings, and other naturally-comparable objects.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference#dimension_group

    Attributes:
    ----------
    :param varName:
    """
    return Heap(selector="%s.heap" % self.varId, varName=varName, setVar=True)


class Group(JsPackage):

  def size(self):
    """
    Description:
    -----------
    Returns the number of distinct values in the group, independent of any filters; the cardinality.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference
    """
    return JsObjects.JsNumber.JsNumber("%s.size()" % self.varId)

  def reduce(self, add, remove, initial):
    """
    Description:
    -----------
    Specifies the reduce functions for this grouping, and returns this grouping.
    The default behavior, reduce by count, is implemented as follows

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param add:
    :param remove:
    :param initial:
    """
    raise Exception("Not implemented yet !")

  def reduceCount(self, value):
    """
    Description:
    -----------
    A convenience method for setting the reduce functions to count records; returns this group.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference#dimension_group

    Attributes:
    ----------
    :param value: String. The column name

    :return: returns this group
    """
    return self.fnc("reduceCount(function(d) { return d['%s'] ;})" % value)

  def reduceSum(self, value):
    """
    Description:
    -----------
    A convenience method for setting the reduce functions to sum records using the specified value accessor function;

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param value: String. The column name

    :return: returns this group
    """
    return self.fnc("reduceSum(function(d) { return d['%s'] ;})" % value)

  def order(self, orderValue):
    """
    Description:
    -----------

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    :param orderValue:
    """
    raise Exception("Not implemented yet !")

  def orderNatural(self):
    """
    A convenience method for using natural order for reduce values. Returns this grouping

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference
    """
    groupObj = Group("%s.orderNatural()" % self.toStr())
    return groupObj

  def top(self, k=None):
    """
    Description:
    -----------
    Returns a new array containing the top k groups, according to the group order of the associated reduce value.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    :param k:
    """
    if k is None:
      return JsObjects.JsArray.JsArray("%s.top()" % self.varId)

    return JsObjects.JsArray.JsArray("%s.top(%s)" % (self.varId, k))

  def all(self):
    """
    Description:
    -----------
    Returns the array of all groups, in ascending natural order by key. Like top, the returned objects contain key and value attributes.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference
    """
    return JsObjects.JsArray.JsArray("%s.all()" % self.varId)

  def dispose(self):
    """
    Description:
    -----------
    Removes this group from its dimension.
    This group will no longer update when new filters are applied to the crossfilter, and it may be garbage collected if there are no other references to it remaining.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference
    """
    return JsObjects.JsObject.JsObject("%s.dispose()" % self.toStr())


class GroupAll(JsPackage):

  def reduceCount(self, value):
    """
    Description:
    -----------
    A convenience method for setting the reduce functions to count records; returns this group.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference#dimension_group

    Attributes:
    ----------
    :param value: String. The column name

    :return: returns this group
    """
    return self.fnc("reduceCount(function(d) { return d['%s'] ;})" % value)

  def reduceSum(self, value):
    """
    Description:
    -----------
    A convenience method for setting the reduce functions to sum records using the specified value accessor function;

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference

    Attributes:
    ----------
    :param value: String. The column name

    :return: returns this group
    """
    return self.fnc("reduceSum(function(d) { return d['%s'] ;})" % value)

  def value(self):
    """
    Description:
    -----------
    Equivalent to group.all()[0].value.

    Related Pages:

			https://github.com/crossfilter/crossfilter/wiki/API-Reference
    """
    return JsObjects.JsNumber.JsNumber("%s.value()" % self.varId)





import json

from epyk.core.js import JsUtils
from epyk.core.js.html import JsHtml
from epyk.core.js.objects import JsNodeDom

from epyk.core.html import Defaults
from epyk.core.js.primitives import JsObjects


class JsShapes(object):
  shapes = ['triangle', 'rectangle', 'circle', 'arrow']

  def _svg(self, shape_def):
    return '''
        var width = options.svg_style.width; var height = options.svg_style.height;
        var svgns = '%(svgns)s';
        var svg = document.createElementNS(svgns, 'svg');
        var defs = document.createElementNS(svgns, 'defs');
        var gradient = document.createElementNS(svgns, 'linearGradient');
        
        for (var i = 0, length = options.colors[step.status].length; i < length; i++) {
          var stop = document.createElementNS(svgns, 'stop');
          stop.setAttribute('offset', options.colors[step.status][i].offset);
          stop.setAttribute('stop-color', options.colors[step.status][i].color);
          gradient.appendChild(stop)}
        
        gradient.id = 'gradient' + step.status + htmlObj.id; gradient.setAttribute('x1', '0'); gradient.setAttribute('x2', '0'); 
        gradient.setAttribute('y1', '0'); gradient.setAttribute('y2', '1');
        defs.appendChild(gradient);
        %(shape_def)s;
        shape.setAttribute('fill', 'url(#'+ gradient.id +')');
        shape.setAttribute('name', 'signal');
        svg.setAttribute('width', width +'px');
        svg.setAttribute('height', height +'px')
        svg.setAttribute('version', '1.1');
        svg.setAttribute('xmlns', svgns);
        
        if(options.line){
          var line = document.createElementNS(svgns, 'line');
          line.setAttribute('x1', 0);
          line.setAttribute('x2', width);
          line.setAttribute('y1', height/2 );
          line.setAttribute('y2', height/2 );
          for (var key in options.line){ line.setAttribute(key, options.line[key]) }
          svg.appendChild(line);
        }
        svg.appendChild(defs); svg.appendChild(shape); htmlObj.insertBefore(svg, htmlObj.firstChild);
        htmlObj.setAttribute('data-status', step.status);
      ''' % {'shape_def': shape_def, 'svgns': Defaults.SVGNS}

  def triangle(self):
    shape_def = '''
      var shape = document.createElementNS(svgns, 'polygon');
      shape.setAttribute('points', ''+ ((width)/2) + ' 0,'+ (width - 20) +' '+ height +',20 '+ height);
      shape.setAttribute('stroke', options.backgrounds[step.status]);
    '''
    return self._svg(shape_def)

  def rectangle(self):
    shape_def = '''
      var shape = document.createElementNS(svgns, 'rect');
      shape.setAttribute('stroke', options.backgrounds[step.status]);
      shape.setAttribute('width', width-20);
      shape.setAttribute('x', 10);
      shape.setAttribute('height', height);
    '''
    return self._svg(shape_def)

  def circle(self):
    shape_def = '''
      var shape = document.createElementNS(svgns, 'circle');
      shape.setAttribute('stroke', options.backgrounds[step.status]);
      shape.setAttribute('stroke-width', 1);
      shape.setAttribute('cx', width / options.circle_factor );
      shape.setAttribute('cy', height / options.circle_factor );
      shape.setAttribute('r', height / options.circle_factor );
    '''
    return self._svg(shape_def)

  def arrow(self):
    shape_def = '''
      var shape = document.createElementNS(svgns, 'polygon');
      shape.setAttribute('points', '10 0,'+ (width-30) +' 0,'+ (width-10) + ' ' + (height/2) +','+ (width-30) +' '+ height +',10 0'+ height +',25 '+ (height/2) +',10 0');
      shape.setAttribute('stroke', options.backgrounds[step.status]);
    '''
    return self._svg(shape_def)

  def custom(self, shape_def):
    return self._svg(shape_def)


class Step(JsNodeDom.JsDoms):

  def __init__(self, data, varName=None, setVar=False, isPyData=False, report=None, src=None):
    super(Step, self).__init__(data, varName, setVar, isPyData, report)
    self._src = src

  def colors(self, colors, status='bespoke', clear_gradient=True):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param colors:
    """
    clear_gradient = JsUtils.jsConvertData(clear_gradient, None)
    if not isinstance(colors[0], dict):
      s = 100 / (len(colors) - 1)
      tmp_colors = []
      for i, c in enumerate(colors):
        tmp_colors.append({"color": c, 'offset': "%s%%" % int(i * s)})
      colors = JsUtils.jsConvertData(tmp_colors, None)
    return JsObjects.JsObjects.get('''
      var gradient = document.createElementNS(%(svgns)s, 'linearGradient');
      var gradient_colors = %(colors)s;
      for (var i = 0, length = gradient_colors.length; i < length; i++) {
          var stop = document.createElementNS(%(svgns)s, 'stop');
          stop.setAttribute('offset', gradient_colors[i].offset);
          stop.setAttribute('stop-color', gradient_colors[i].color);
          gradient.appendChild(stop)}
      gradient.id = 'gradient_%(status)s'; gradient.setAttribute('x1', '0'); gradient.setAttribute('x2', '0'); 
      gradient.setAttribute('y1', '0'); gradient.setAttribute('y2', '1');
      if(%(clear_gradient)s){%(comp)s.querySelector('defs').innerHTML = '';};
      %(comp)s.querySelector('defs').appendChild(gradient);
      %(comp)s.querySelector('[name=signal]').setAttribute('fill', 'url(#'+ gradient.id +')');
      %(comp)s.querySelector('[name=signal]').setAttribute('stroke', gradient_colors[0]["color"]);
      %(comp)s.setAttribute('data-status', '%(status)s');
      ''' % {"colors": colors, 'comp': self.varName, 'svgns': JsUtils.jsConvertData(Defaults.SVGNS, None),
             'clear_gradient': clear_gradient, 'status': status})

  def success(self):
    return self.colors(self._src.options.success, status='success')

  def error(self):
    return self.colors(self._src.options.error, status='error')

  def pending(self):
    return self.colors(self._src.options.pending, status='pending')

  def waiting(self):
    return self.colors(self._src.options.waiting, status='waiting')

  def blink(self):
    """
    Description:
    ------------

    """

  @property
  def status(self):
    return JsObjects.JsObjects.get('%s.getAttribute("data-status")' % self.varName)

  def shape(self, shape, status='success', step=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param shape:
    :param status:
    """
    step = {"status": status} if step is None else JsUtils.jsConvertData(step, None)
    return JsObjects.JsObjects.get('''
      var svgns = '%(svgns)s';
      %(comp)s.querySelector('svg').remove(); %(shape)s(%(comp)s, %(options)s, %(step)s)
      ''' % {'svgns': Defaults.SVGNS, 'comp': self.varName, 'options': json.dumps(self._src._jsStyles), 'step': step, 'shape': shape})

  def triangle(self, status='error', step=None):
    """
    Description:
    ------------
    Hide all the panels in the drawer component
    """
    step = {"status": status} if step is None else JsUtils.jsConvertData(step, None)
    return JsObjects.JsObjects.get('''
      %(comp)s.querySelector('svg').remove(); triangle(%(comp)s, %(options)s, %(step)s)
      ''' % {'comp': self.varName, 'options': json.dumps(self._src._jsStyles), 'step': step})

  def rectangle(self, status='success', step=None):
    """
    Description:
    ------------
    Hide all the panels in the drawer component

    htmlObj.querySelector('svg').remove();
    """
    step = {"status": status} if step is None else JsUtils.jsConvertData(step, None)
    return JsObjects.JsObjects.get('''
      %(comp)s.querySelector('svg').remove(); rectangle(%(comp)s, %(options)s, %(step)s)
      ''' % {'comp': self.varName, 'options': json.dumps(self._src._jsStyles), 'step': step})

  def arrow(self, status='success', step=None):
    """
    Description:
    ------------
    Hide all the panels in the drawer component
    """
    step = {"status": status} if step is None else JsUtils.jsConvertData(step, None)
    return JsObjects.JsObjects.get('''
      %(comp)s.querySelector('svg').remove(); arrow(%(comp)s, %(options)s, %(step)s)
      ''' % {'comp': self.varName, 'options': json.dumps(self._src._jsStyles), 'step': step})

  def circle(self, status='waiting', step=None):
    """
    Description:
    ------------
    Hide all the panels in the drawer component
    """
    step = {"status": status} if step is None else JsUtils.jsConvertData(step, None)
    return JsObjects.JsObjects.get('''
      %(comp)s.querySelector('svg').remove(); circle(%(comp)s, %(options)s, %(step)s)
      ''' % {'comp': self.varName, 'options': json.dumps(self._src._jsStyles), 'step': step})

  def label(self, value):
    """
    Description:
    ------------
    Add a text label below the shape

    Attributes:
    ----------
    :param value: String. The text to be added
    """
    return self.querySelector('span[name=label]').innerHTML(value)

  def text(self, text, x=None, y=None, css=None):
    """
    Description:
    ------------
    Add a text on the shape

    Attributes:
    ----------
    :param text: String. The text to be added
    :param x: Number. Optional. The x position
    :param y: Number. Optional. The y position
    :param css: Dictionary. Optional. The CSS attributes
    """
    text = JsUtils.jsConvertData(text, None)
    if y is None:
      y = self._src._jsStyles['svg_style']['height'] / 2 + 3
    if x is None:
      x = self._src._jsStyles['svg_style']['width'] / 2
    dft_css = {"text-anchor": "middle", 'fill': 'white'}
    if css is not None:
      dft_css.update(css)
    return JsObjects.JsObjects.get('''
      var newText = document.createElementNS("http://www.w3.org/2000/svg", "text");
      newText.setAttributeNS(null, 'x', %(x)s);
      newText.setAttributeNS(null, 'y', %(y)s); var css = %(css)s;
      for (var key in css){ newText.setAttributeNS(null, key, css[key]) }
      var textNode = document.createTextNode(%(text)s);
      newText.appendChild(textNode);
      %(svg)s.appendChild(newText)''' % {'text': text, 'svg': self.querySelector('svg'), 'y': y, 'x': x, 'css': dft_css})

  def blink(self, color="#FFDEB3", border="#FF9200"):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param color:
    :param border:
    """
    return JsObjects.JsObjects.get('''
      var svg_holder = %(comp)s; 
      if (svg_holder.getAttribute("data-blink") === null){
        svg_holder.setAttribute("data-blink", true);
        if (svg_holder.getAttribute("data-gradient") === null){
          svg_holder.setAttribute("data-gradient", svg_holder.querySelector('[name=signal]').getAttribute('fill') );
          svg_holder.setAttribute("data-border", svg_holder.querySelector('[name=signal]').getAttribute('stroke') );
        };
        %(colors)s
      } else {
        svg_holder.removeAttribute("data-blink");
        svg_holder.querySelector('[name=signal]').setAttribute("fill", svg_holder.getAttribute("data-gradient"));
        svg_holder.querySelector('[name=signal]').setAttribute("stroke", svg_holder.getAttribute("data-border"));
      }
      ''' % {"comp": self.varName, 'colors': self.colors([color, border], clear_gradient=False)})

  def click(self, jsFncs):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param jsFncs: Array the Javascript actions when click is triggered
    """
    return JsObjects.JsObjects.get('%s.addEventListener("click", function(){%s})' % (self.varName, JsUtils.jsConvertFncs(jsFncs, toStr=True)))


class Stepper(JsHtml.JsHtmlRich):

  def _get_index(self):
    constructors = self._report._props.setdefault("js", {}).setdefault("constructors", {})
    constructors['StepByName'] = "function StepByName(doms, label){var index_svg = -1; doms.childNodes.forEach(function(dom, i){if(dom.querySelector('[name=label]').innerHTML == label){ index_svg = i}}); return index_svg}"
    return 'StepByName'

  @property
  def val(self):
    """
    Description:
    -----------

    """
    return JsObjects.JsObjects.get("{%s: {value: %s, timestamp: Date.now(), offset: new Date().getTimezoneOffset()} }" % (
      self.htmlCode, self.content.toStr()))

  @property
  def content(self):
    return JsHtml.ContentFormatters(self._report, "%s.innerHTML" % self.varName)

  def __getitem__(self, i):
    return Step(self, src=self._src, varName=self.querySelectorAll("div[name=svg_holder]")[i], report=self._report)

  def getByLabel(self, label):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param label:
    """
    label = JsUtils.jsConvertData(label, None)
    return Step(self, src=self._src, varName='''
      %(holder)s.querySelectorAll('div[name=svg_holder]')[%(StepByName)s(%(holder)s, %(label)s)]''' % {'StepByName': self._get_index(), "holder": self.varName, 'label': label}, report=self._report)

  def addItem(self, label, tooltip="", shape='circle',  status="waiting"):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param label:
    :param tooltip:
    :param shape:
    :param status:
    """
    tooltip = JsUtils.jsConvertData(tooltip, None)
    label = JsUtils.jsConvertData(label, None)
    shape = JsUtils.jsConvertData(shape, None)
    status = JsUtils.jsConvertData(status, None)
    return '''
      if(%(StepByName)s(%(holder)s, %(label)s) == -1){
        var li = document.createElement("LI") ; var options = %(options)s;
        var span = document.createElement("SPAN"); span.setAttribute('name', 'label'); span.innerHTML = %(label)s;
        var div = document.createElement("DIV"); 
        div.setAttribute('title', %(tooltip)s);
        div.setAttribute('name', 'svg_holder');
        for (var key in options.text_style){ span.style[key] = options.text_style[key] }
        span.style.width = options.svg_style.width +"px";
  
        div.appendChild(span); li.appendChild(div);
        div.id = "svg_" + %(holder)s.childElementCount;
        window[%(shape)s](div, options, %(status)s); 
        %(holder)s.appendChild(li)}
      ''' % {'StepByName': self._get_index(), "holder": self.varName, 'status': status, 'shape': shape, "tooltip": tooltip, 'label': label, 'options': json.dumps(self._src._jsStyles)}

  def delete(self, i):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param i:
    """
    return self[i].remove()


class Drawer(JsHtml.JsHtmlRich):

  @property
  def content(self):
    """
    Description:
    ------------
    Hide all the panels in the drawer component
    """
    return JsHtml.ContentFormatters(self._report, ''' 
      (function(doms, contents){var index =-1; doms.childNodes.forEach(function(dom, k){if(dom.style.display !== 'none'){index = k}}); 
        if (index >= 0){return contents.childNodes[index].innerHTML} else{return index}  })(%s, %s)
        ''' % (self.querySelector("div[name=drawer_panels]"), self.querySelector("div[name=drawer_content]")))

  def hide(self):
    """
    Description:
    ------------
    Hide all the panels in the drawer component
    """
    return JsObjects.JsObjects.get(''' 
      (function(doms){doms.childNodes.forEach(function(dom){dom.style.display = 'none'; })})(%s)
      ''' % self.querySelector("div[name=drawer_panels]"))

  def add(self, link, panel=""):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param link:
    :param panel:
    """
    return JsObjects.JsObjects.get('''
      ''' % self.querySelectorAll("[name=drawer_panels]"))

  def delete(self, i):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param link:
    """
    return JsObjects.JsObjects.get(''' 
          %(panel)s.childNodes[%(i)s].remove(); %(drawer)s.childNodes[%(i)s].remove(); 
          ''' % {'panel': self.querySelector("div[name=drawer_panels]"), 'drawer': self.querySelector("div[name=drawer_content]"), 'i': i})




from epyk.core.html.options import Options


class OptionsSlider(Options):

  @property
  def animate(self):
    """
    Description:
    ------------
    Whether to slide the handle smoothly when the user clicks on the slider track. Also accepts any valid animation

    Related Pages:

			https://api.jqueryui.com/slider/#option-animate

		:param animate:
    """
    return self._config_get(False)

  @animate.setter
  def animate(self, value):
    self._config(value)

  @property
  def classes(self):
    """
    Description:
    ------------
    Specify additional classes to add to the widget's elements.
    Any of classes specified in the Theming section can be used as keys to override their value.
    To learn more about this option, check out the learn article about the classes option.

    Related Pages:

			https://api.jqueryui.com/slider/#option-classes

		:param classes:
    """
    return self._config_get('classes')

  @classes.setter
  def classes(self, value):
    self._config(value)

  @property
  def disabled(self):
    """
    Description:
    ------------
    Disables the slider if set to true.

    Related Pages:

			https://api.jqueryui.com/slider/#option-disabled

		:param disabled:
    """
    return self._config_get(False)

  @disabled.setter
  def disabled(self, value):
    self._config(value)

  @property
  def max(self):
    """
    Description:
    ------------
    The maximum value of the slider.

    Related Pages:

			https://api.jqueryui.com/slider/#option-max

		:param max:
    """
    return self._config_get(100)

  @max.setter
  def max(self, value):
    self._config(value)

  @property
  def min(self):
    """
    Description:
    ------------
    The minimum value of the slider.

    Related Pages:

			https://api.jqueryui.com/slider/#option-min

		:param min:
    """
    return self._config_get(0)

  @min.setter
  def min(self, value):
    self._config(value)

  @property
  def orientation(self):
    """
    Description:
    ------------
    Determines whether the slider handles move horizontally (min on left, max on right) or vertically (min on bottom, max on top).
    Possible values: "horizontal", "vertical".

    Related Pages:

			https://api.jqueryui.com/slider/#option-orientation
    """
    return self._config_get("horizontal")

  @orientation.setter
  def orientation(self, value):
    self._config(value)

  @property
  def range(self):
    """
    Description:
    ------------
    Whether the slider represents a range.

    Related Pages:

			https://api.jqueryui.com/slider/#option-range
    """
    return self._config_get(False)

  @range.setter
  def range(self, value):
    self._config(value)

  @property
  def step(self):
    """
    Description:
    ------------
    Determines the size or amount of each interval or step the slider takes between the min and max.
    The full specified value range of the slider (max - min) should be evenly divisible by the step.

    Related Pages:

			https://api.jqueryui.com/slider/#option-step
    """
    return self._config_get(1)

  @step.setter
  def step(self, value):
    self._config(value)

  @property
  def value(self):
    """
    Description:
    ------------
    Determines the value of the slider, if there's only one handle.
    If there is more than one handle, determines the value of the first handle.

    Related Pages:

			https://api.jqueryui.com/slider/#option-value
    """
    return self._config_get(0)

  @value.setter
  def value(self, value):
    self._config(value)

  @property
  def values(self):
    """
    Description:
    ------------
    This option can be used to specify multiple handles.
    If the range option is set to true, the length of values should be 2.

    Related Pages:

			https://api.jqueryui.com/slider/#option-values
    """
    return self._config_get(0)

  @values.setter
  def values(self, value):
    self._config(value)


class OptionsProgBar(Options):

  @property
  def classes(self):
    """
    Description:
    ------------
    Initialize the progressbar with the classes option specified, changing the theming for the ui-progressbar

    Related Pages:

			https://api.jqueryui.com/progressbar/#option-classes
    """
    return self._config_get(None)

  @classes.setter
  def classes(self, value):
    self._config(value)

  @property
  def background(self):
    """
    Description:
    ------------

    """
    return self._config_group_get('css', None)

  @background.setter
  def background(self, value):
    self._config_group('css', value)

  def css(self, attrs):
    """
    Description:
    ------------

    :param attrs:
    """
    css_attrs = self._config_get('css', {})
    css_attrs.update(attrs)
    self._config(css_attrs)

  @property
  def disabled(self):
    """
    Description:
    ------------
    Disables the progressbar if set to true.

    Related Pages:

			https://api.jqueryui.com/progressbar/#option-disabled
    """
    return self._config_get(None)

  @disabled.setter
  def disabled(self, bool):
    self._config(bool)

  @property
  def max(self):
    """
    Description:
    ------------
    The maximum value of the progressbar.

    Related Pages:

			https://api.jqueryui.com/progressbar/#option-max
    """
    return self._config_get(100)

  @max.setter
  def max(self, num):
    self._config(num)

  @property
  def value(self):
    """
    Description:
    ------------
    The value of the progressbar.

    Related Pages:

			https://api.jqueryui.com/progressbar/#option-value
    """
    return self._config_get(None)

  @value.setter
  def value(self, val):
    self._config(val)


class OptionsMenu(Options):

  @property
  def classes(self):
    """
    Description:
    ------------
    Specify additional classes to add to the widget's elements.
    Any of classes specified in the Theming section can be used as keys to override their value.
    To learn more about this option, check out the learn article about the

    Related Pages:

			https://api.jqueryui.com/menu/#option-classes
    """
    return self._config_get({})

  @classes.setter
  def classes(self, value):
    self._config(value)

  @property
  def disabled(self):
    """
    Description:
    ------------
    Disables the menu if set to true

    Related Pages:

			https://api.jqueryui.com/menu/#option-disabled
    """
    return self._config_get(False)

  @disabled.setter
  def disabled(self, bool):
    self._config(bool)

  @property
  def icons(self):
    """
    Description:
    ------------
    Icons to use for submenus, matching an icon provided by the jQuery UI CSS Framework.

    Related Pages:

			https://api.jqueryui.com/menu/#option-icons
    """
    return self._config_get(None)

  @icons.setter
  def icons(self, bool):
    self._config(bool)

  @property
  def position(self):
    """
    Description:
    ------------
    Identifies the position of submenus in relation to the associated parent menu item.
    The of option defaults to the parent menu item, but you can specify another element to position against.
    You can refer to the jQuery UI Position utility for more details about the various options.

    Related Pages:

			https://api.jqueryui.com/menu/#option-position
    """
    return self._config_get(None)

  @position.setter
  def position(self, bool):
    self._config(bool)


class OptionDialog(Options):

  @property
  def appendTo(self):
    """
    Description:
    ------------
    Which element the dialog (and overlay, if modal) should be appended to

    Related Pages:

			https://api.jqueryui.com/dialog/#option-appendTo
    """
    return self._config_get('body')

  @appendTo.setter
  def appendTo(self, bool):
    self._config(bool)

  @property
  def autoOpen(self):
    """
    Description:
    ------------
    If set to true, the dialog will automatically open upon initialization.
    If false, the dialog will stay hidden until the open() method is called.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-autoOpen
    """
    return self._config_get(True)

  @autoOpen.setter
  def autoOpen(self, bool):
    self._config(bool)

  @property
  def classes(self):
    """
    Description:
    ------------
    Specify additional classes to add to the widget's elements.
    Any of classes specified in the Theming section can be used as keys to override their value.
    To learn more about this option, check out the learn article about the

    Related Pages:

			https://api.jqueryui.com/dialog/#option-classes
    """
    return self._config_get({})

  @classes.setter
  def classes(self, value):
    self._config(value)

  @property
  def closeOnEscape(self):
    """
    Description:
    ------------
    Specifies whether the dialog should close when it has focus and the user presses the escape (ESC) key.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-closeOnEscape
    """
    return self._config_get(True)

  @closeOnEscape.setter
  def closeOnEscape(self, value):
    self._config(value)

  @property
  def closeText(self):
    """
    Description:
    ------------
    Specifies the text for the close button. Note that the close text is visibly hidden when using a standard theme.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-closeText
    """
    return self._config_get(True)

  @closeText.setter
  def closeText(self, value):
    self._config(value)

  @property
  def draggable(self):
    """
    Description:
    ------------
    If set to true, the dialog will be draggable by the title bar. Requires the jQuery UI Draggable widget to be included.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-draggable
    """
    return self._config_get(True)

  @draggable.setter
  def draggable(self, value):
    self._config(value)

  @property
  def height(self):
    """
    Description:
    ------------
    The height of the dialog.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-height
    """
    return self._config_get('auto')

  @height.setter
  def height(self, value):
    self._config(value)

  @property
  def hide(self):
    """
    Description:
    ------------
    If and how to animate the hiding of the dialog.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-hide
    """
    return self._config_get(None)

  @hide.setter
  def hide(self, value):
    self._config(value)

  @property
  def maxHeight(self):
    """
    Description:
    ------------
    The maximum height to which the dialog can be resized, in pixels.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-maxHeight
    """
    return self._config_group_get("option", None)

  @maxHeight.setter
  def maxHeight(self, value):
    self._config_group("option", value)

  @property
  def maxWidth(self):
    """
    Description:
    ------------
    The maximum width to which the dialog can be resized, in pixels.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-maxWidth
    """
    return self._config_group_get("option", None)

  @maxWidth.setter
  def maxWidth(self, value):
    self._config_group("option", value)

  @property
  def minHeight(self):
    """
    Description:
    ------------
    The minimum height to which the dialog can be resized, in pixels.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-minHeight
    """
    return self._config_group_get("option", None)

  @minHeight.setter
  def minHeight(self, value):
    self._config_group("option", value)

  @property
  def minWidth(self):
    """
    Description:
    ------------
    The minimum width to which the dialog can be resized, in pixels.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-minWidth
    """
    return self._config_group_get("option", None)

  @minWidth.setter
  def minWidth(self, value):
    self._config_group("option", value)

  def position(self, my="center", at="center", of="window"):
    """
    Description:
    ------------
    Specifies where the dialog should be displayed when opened. The dialog will handle collisions such that as much of the dialog is visible as possible.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-position

    :param my: String.
    :param at: String.
    :param of: String.
    """
    self._config({"my": my, "at": at, "of": of})
    return self

  @property
  def modal(self):
    """
    Description:
    ------------
    If set to true, the dialog will have modal behavior; other items on the page will be disabled, i.e., cannot be interacted with.
    Modal dialogs create an overlay below the dialog but above other page elements.

    Related Pages:

      https://api.jqueryui.com/dialog/#option-modal
    """
    return self._config_group_get("option", None)

  @modal.setter
  def modal(self, value):
    self._config_group("option", value)

  @property
  def resizable(self):
    """
    Description:
    ------------
    If set to true, the dialog will be resizable. Requires the jQuery UI Resizable widget to be included.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-resizable
    """
    return self._config_group_get("option", None)

  @resizable.setter
  def resizable(self, value):
    self._config_group("option", value)

  @property
  def title(self):
    """
    Description:
    ------------
    Specifies the title of the dialog. If the value is null, the title attribute on the dialog source element will be used.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-title
    """
    return self._config_get(None)

  @title.setter
  def title(self, value):
    self._config(value)

  @property
  def width(self):
    """
    Description:
    ------------
    The width of the dialog, in pixels.

    Related Pages:

			https://api.jqueryui.com/dialog/#option-title

		:param width:
    """
    return self._config_get(300)

  @width.setter
  def width(self, value):
    self._config(value)


class OptionBar(Options):

  @property
  def draggable(self):
    """
    Description:
    ------------
    Set the bar draggable using Jquery UI
    """
    return self._config_get(False)

  @draggable.setter
  def draggable(self, bool):
    self._config(bool)




from epyk.core.html import Html

from epyk.core.js import JsUtils
from epyk.core.js.packages import JsGoogleAPI

from epyk.core.html.options import OptGoogle


class ChartGeoGoogle(Html.Html):
  name = 'Google Chart'
  requirements = ('google-maps', )

  def __init__(self,  report, width, height, options, htmlCode, profile):
    super(ChartGeoGoogle, self).__init__(report, [], htmlCode=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self.style.css.margin = "10px 0"
    self.__options = OptGoogle.OptionMaps(self, options)

  @property
  def chartId(self):
    """
    Description:
    -----------
    Return the Javascript variable of the chart
    """
    return "%s_obj" % self.htmlCode

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript base function

    Return all the Javascript functions defined in the framework.
    THis is an entry point to the full Javascript ecosystem.

    :return: A Javascript object

    :rtype: JsChartJs.JsChart
    """
    if self._js is None:
      self._js = JsGoogleAPI.GoogleMapsAPI(selector="window['%s']" % self.chartId, src=self)
    return self._js

  @property
  def options(self):
    """
    Description:
    -----------

    :rtype: OptGoogle.OptionMaps
    """
    return self.__options

  def build(self, data=None, options=None, profile=False):

    js_options = []
    for k, v in self._jsStyles.items():
      if self.options.isJsContent(k) or str(v).strip().startswith("function"):
        js_options.append("%s: %s" % (k, v))
      else:
        js_options.append("%s: %s" % (k, JsUtils.jsConvertData(v, None)))
    return '%s = new google.maps.Map(%s, {%s})' % (self.chartId, self.dom.varId, ",".join(js_options))

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %s></div>' % self.get_attrs(pyClassNames=self.style.get_classes())




from epyk.core.js import JsUtils


class JsFor(object):

  def __init__(self, end, options=None):
    """
    Description:
    -----------
    The for statement creates a loop that is executed as long as a condition is true.

    The loop will continue to run as long as the condition is true. It will only stop when the condition becomes false.

    Related Pages:

			https//www.w3schools.com/jsref/jsref_for.asp
    https://www.w3schools.com/js/js_performance.asp

    Attributes:
    ----------
    :param end:
    :param options: Dictionary
    """
    self.options = {"var": 'i', 'start': 0, 'step': 1, 'end': end}
    if options is not None:
      self.options.update(options)
    self.__jsFncs = []

  @property
  def var(self):
    """

    :return:
    """
    return self.options['var']

  @var.setter
  def var(self, value):
    """

    :return:
    """
    self.options['var'] = value

  @property
  def start(self):
    """

    :return:
    """
    return self.options['start']

  @start.setter
  def start(self, value):
    """

    :return:
    """
    self.options['start'] = value

  @property
  def end(self):
    """

    :return:
    """
    return self.options['end']

  @end.setter
  def end(self, value):
    """

    :return:
    """
    self.options['end'] = value

  @property
  def step(self):
    """

    :return:
    """
    return self.options['step']

  @step.setter
  def step(self, value):
    """

    :return:
    """
    self.options['step'] = value

  def fncs(self, jsFncs, reset=True):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFncs:
    :param reset:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    if reset:
      self.__jsFncs = jsFncs
    else:
      self.__jsFncs.extend(jsFncs)
    return self

  def toStr(self):
    """

    """
    self.options['expr'] = JsUtils.jsConvertFncs(self.__jsFncs, toStr=True)
    return "for(var %(var)s = %(start)s; %(var)s < %(end)s; %(var)s++){%(expr)s}" % self.options


class JsIterable(object):

  def __init__(self, jsIterable, options=None):
    """

    Attributes:
    ----------
    :param jsIterable:
    """
    self.__js_it = jsIterable
    self.options = {"var": 'x', 'type': 'in'}
    if options is not None:
      self.options.update(options)

  @property
  def var(self):
    """
    Description:
    -----------
`   Return the variable reference for this loop
    """
    return self.options['var']

  @var.setter
  def var(self, value):
    """
    Description:
    -----------
`   Return the variable reference for this loop

    Attributes:
    ----------
    :param value: String. The value reference for the JavaScript variable
    """
    self.options['var'] = value

  def fncs(self, jsFncs, reset=True):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFncs:
    :param reset:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    if reset:
      self.__jsFncs = jsFncs
    else:
      self.__jsFncs.extend(jsFncs)
    return self

  def toStr(self):
    """

    """
    jsNfncs = JsUtils.jsConvertFncs(self.__jsFncs, toStr=True)
    jsIter = JsUtils.jsConvertData(self.__js_it, None)
    return "for(var %s %s %s){%s}" % (self.var, self.options['type'], jsIter, jsNfncs)
